--***************************************************************************
-- DBTSV.E - TSV Database 1.1
-- Created by Andrs Szab (salix@freemail.hu) 2006-04-30 13:33:57
--***************************************************************************
-- This is the include file for a very simple TSV database. The TSV (Tab
-- Separated Values) has been officially registered as an Internet media
-- type (MIME type) under the name "text/tab-separated-values". TSV is one
-- way of making tabular data available on a network in an universally
-- accessible way, though it may require some skills on the user side to
-- know what could be done with the data and do something useful with it.
-- 	* A file in TSV format consists of lines.
-- 	* Each line contain fields separated from each other by TAB
--		characters (horizontal tab, HT, Ascii control code 9, \t).
-- 	* "Field" means here just any string of characters, excluding TABs.
-- 		The point is simply that TABs divide a line into pieces,
--		components.
-- 	* Each line must contain the same number of fields.
-- 	* The first line contains the names for the fields.
-- There is more about TSV at http://www.cs.tut.fi/~jkorpela/TSV.html
-- In addition DBTSV.E handles the first field as a key of the record.
--***************************************************************************
-- Database routines:
-- 	tsv_create(sequence tsv_file, sequence tsv_head)
-- 	tsv_open(sequence tsv_file, sequence tsv_mode )
-- 	tsv_close(sequence tsv_file)
-- 	tsv_fields(sequence tsv_file)
-- 	tsv_keys(sequence tsv_file)
-- 	tsv_size(sequence tsv_file)
-- 	tsv_rec_get(sequence tsv_file, sequence tsv_key)
-- 	tsv_rec_filter(sequence tsv_file, object filter_id)
-- 	tsv_rec_all(sequence tsv_file)
-- 	tsv_rec_add(sequence tsv_file, sequence tsv_rec)
-- 	tsv_rec_del(sequence tsv_file, sequence tsv_key)
-- 	tsv_rec_key(sequence tsv_file, sequence tsv_key)
-- 	tsv_rec_nr(sequence tsv_file, integer tsv_line)
-- 	tsv_rec_dump(sequence tsv_file, sequence tsv_recs)
--***************************************************************************
-- Changes:
-- 	V1.1 bug in tsv_size() corrected not to show less number of lines
-- 	V1.1 usage of filter_id in tsv_rec_filter() is modified
-- 	V1.1 handling the last line in tsv_rec_filter() was improved
-- 	V1.1 demo3 had to be corrected as EDS allows only unique keys
-- 	V1.1 demo4 is removed to reduce the pack size
--***************************************************************************
-- Advantages:
--	* handy database files (you can modify/check them a text editor)
--	* easy to use routines
-- Disadvantages:
--	* (very!) slow, specially deleting and modifying records
--	* needs a huge amount of memory, specially deleting and modifying
--	* the stored data should be text
--***************************************************************************

with trace
without warning
without profile

--***************************************************************************
-- Some internal variables, you won't really need them
--***************************************************************************

global constant TSV_ERROR=0,
	TSV_OK=1,
	TSV_ERROR_USED=-1,	-- the file is already in use
	TSV_ERROR_OPEN=-2,	-- couldn't open the file for...
	TSV_ERROR_CLOSE=-3, 	-- couldn't close
	TSV_ERROR_MODE=-4,	-- open but in bad mode
	TSV_ERROR_FIND=-5	-- couldn't find something
global constant READ="r",
	UPDATE="u",
	WRITE="w",
	APPEND="a"
constant TSV_NAME=1,	-- TSV_FILES[i][1] is the name of the file
	TSV_NR=2,	-- TSV_FILES[i][1] is the name of the file
	TSV_MODE=3
constant KEY=1
sequence TSV_FILES	-- list of open TSV files
TSV_FILES={}
integer fn, fc, fx 	-- generally used for file operations
--***************************************************************************

--***************************************************************************
-- Subroutines for the database calls
--***************************************************************************

function tsv_file_check(sequence file_name)
-- this one simply checks if a file is already open
	for i=1 to length(TSV_FILES) by 1 do
		if equal(TSV_FILES[i][TSV_NAME],file_name) then
			return i
		end if
	end for
	return TSV_ERROR
end function

--***************************************************************************
-- These routines open, close, and create TSV files
--***************************************************************************

global function tsv_create(sequence tsv_file, sequence tsv_head)
	-- you can not create a file that is already open
	if tsv_file_check(tsv_file)>0 then
		return TSV_ERROR_USED
	end if
	-- you should be able to open it for writing
	fn=open(tsv_file,WRITE)
	if fn<1 then
		return TSV_ERROR_OPEN
	end if
	-- tsv_head should be a list of sequences/words
	-- tsv_head shouldn't contain \t, \r, \n
	-- these requirement is not checked here (yet)
	TSV_FILES=append(TSV_FILES,{tsv_file,fn,WRITE})
	puts(fn,tsv_head[1])
	for i=2 to length(tsv_head) by 1 do
		puts(fn,"\t"&tsv_head[i])
	end for
	return TSV_OK
end function

global function tsv_open(sequence tsv_file, sequence tsv_mode )
	-- you can not open a file that is already open
	if tsv_file_check(tsv_file)>0 then
		return TSV_ERROR_USED
	end if
	-- you should be able to open it
	fn=open(tsv_file,tsv_mode)
	if fn<1 then
		return TSV_ERROR_OPEN
	end if
	TSV_FILES=append(TSV_FILES,{tsv_file,fn,tsv_mode})
	return TSV_OK
end function

global function tsv_close(sequence tsv_file)
	-- you can not close a file that is not open
	fx=tsv_file_check(tsv_file)
	if fx<1 then
		return TSV_ERROR_CLOSE
	end if
	-- let's close it
	close(TSV_FILES[fx][TSV_NR])
	if fx=1 then
		TSV_FILES=TSV_FILES[2..$]
	elsif fx=length(TSV_FILES) then
		TSV_FILES=TSV_FILES[1..$-1]
	else
		TSV_FILES=TSV_FILES[1..fx-1]&TSV_FILES[fx+1..$]
	end if
	return TSV_OK
end function

--***************************************************************************
-- These routines return basic characteristic of the TSV file
--***************************************************************************

global function tsv_fields(sequence tsv_file)
	sequence sx
	-- you can not read from a file that is not open
	fx=tsv_file_check(tsv_file)
	if fx<1 then
		return TSV_ERROR_OPEN
	end if
	-- it is not wise to read a record if it is not in read mode
	if equal(TSV_FILES[fx][TSV_MODE],READ)=0
	and equal(TSV_FILES[fx][TSV_MODE],UPDATE)=0
	then
		return TSV_ERROR_MODE
	end if
	-- let's start reading
	sx={""}
	fn=TSV_FILES[fx][TSV_NR]
	while 1 do
		fc=getc(fn)
		if fc='\t' then
			sx=append(sx,"")
		elsif fc='\n' then
			exit
		elsif fc=-1 then
			exit
		elsif fc!='\r' then
			sx[$]&=fc
		end if
	end while
	fx=machine_func(19,{fn,0})
	return sx
end function

global function tsv_keys(sequence tsv_file)
	sequence sx,ax
	integer ix
	-- you can not read from a file that is not open
	fx=tsv_file_check(tsv_file)
	if fx<1 then
		return TSV_ERROR_OPEN
	end if
	-- it is not wise to read a record if it is not in read mode
	if equal(TSV_FILES[fx][TSV_MODE],READ)=0 
	and equal(TSV_FILES[fx][TSV_MODE],UPDATE)=0
	then
		return TSV_ERROR_MODE
	end if
	-- let's start reading
	ax={}
	sx=""
	ix='f'	-- not needed line
	fn=TSV_FILES[fx][TSV_NR]
	while 1 do
		fc=getc(fn)
		if fc=-1 then
			exit
		elsif fc='\n' then
			if ix!='f' then
				ax=append(ax,sx)
				sx=""
			end if
			ix='y'
		elsif ix='y' then
			if fc='\t' then
				ix='n'
			elsif fc!='\r' then
				sx&=fc
			end if
		end if
	end while
	ax=append(ax,sx)
	fx=machine_func(19,{fn,0})
	if equal(ax,{}) then
		return TSV_ERROR_FIND
	else
		return ax
	end if
end function

global function tsv_size(sequence tsv_file)
	integer cx
	-- you can not read from a file that is not open
	fx=tsv_file_check(tsv_file)
	if fx<1 then
		return TSV_ERROR_OPEN
	end if
	-- it is not wise to read a record if it is not in read mode
	if equal(TSV_FILES[fx][TSV_MODE],READ)=0
	and equal(TSV_FILES[fx][TSV_MODE],UPDATE)=0
	then
		return TSV_ERROR_MODE
	end if
	-- let's start reading
	cx=0	-- the first line is not needed
	fn=TSV_FILES[fx][TSV_NR]
	while 1 do
		fc=getc(fn)
		if fc='\n' then
			cx+=1
		elsif fc=-1 then
			exit
		end if
	end while
	fx=machine_func(19,{fn,0})
	return cx
end function

--***************************************************************************
-- These routines modify the records of the TSV file
--***************************************************************************

global function tsv_rec_add(sequence tsv_file, sequence tsv_rec)
	sequence sx
	-- you can not add to a file that is not open
	fx=tsv_file_check(tsv_file)
	if fx<1 then
		return TSV_ERROR_OPEN
	end if
	-- it is not wise to add a record if it is not in append mode
	if equal(TSV_FILES[fx][TSV_MODE],APPEND)=0
	and equal(TSV_FILES[fx][TSV_MODE],WRITE)=0 then
		return TSV_ERROR_MODE
	end if
	-- tsv_rec should be a list of sequences/words
	-- tsv_rec should be a unique key in the file
	-- tsv_rec should be the same length as tsv_head was
	-- tsv_rec shouldn't contain \t, \r, \n
	-- these requirement is not checked here (yet)
	sx="\n"&tsv_rec[1]
	for i=2 to length(tsv_rec) by 1 do
		sx&="\t"&tsv_rec[i]
	end for
	fn=TSV_FILES[fx][TSV_NR]
	puts(fn,sx)
	return TSV_OK
end function

global function tsv_rec_dump(sequence tsv_file, sequence tsv_recs)
	sequence sx
	-- you can not add to a file that is not open
	fx=tsv_file_check(tsv_file)
	if fx<1 then
		return TSV_ERROR_OPEN
	end if
	-- it is not wise to add a record if it is not in append mode
	if equal(TSV_FILES[fx][TSV_MODE],APPEND)=0
	and equal(TSV_FILES[fx][TSV_MODE],WRITE)=0 then
		return TSV_ERROR_MODE
	end if
	-- tsv_recs should be a list of sequences/words
	-- tsv_recs should be a unique key in the file
	-- tsv_recs should be the same length as tsv_head was
	-- tsv_recs shouldn't contain \t, \r, \n
	-- these requirement is not checked here (yet)
	fn=TSV_FILES[fx][TSV_NR]
	for l=1 to length(tsv_recs) do
		sx="\n"&tsv_recs[l][1]
		for i=2 to length(tsv_recs[l]) by 1 do
			sx&="\t"&tsv_recs[l][i]
		end for
		puts(fn,sx)
	end for
	return TSV_OK
end function

-- Remember: deleting is slow and unstable so try to avoid it
-- There is a place for improving; would be nice to KILL undesired bytes
global function tsv_rec_del(sequence tsv_file, sequence tsv_key)
	sequence sx
	integer ix,nx,lx
	-- you can not read from a file that is not open
	fx=tsv_file_check(tsv_file)
	if fx<1 then
		return TSV_ERROR_OPEN
	end if
	-- mode doesn't matter cos I need to close and reopen anyway
	-- let's start reading
	sx=""
	fn=TSV_FILES[fx][TSV_NR]
	nx=open("xxx.tsv","w")
	lx=length(tsv_key)+1	-- +1 is the \t
	ix=TSV_ERROR_FIND
	while 1 do
		fc=getc(fn)
		if fc=-1 then
			if equal(sx[1..lx],tsv_key&'\t')!=1 then
				puts(nx,sx)
				ix=machine_func(20,fn)
			end if
			exit
		elsif fc='\n' then
			sx&='\n'
			if equal(sx[1..lx],tsv_key&'\t')!=1 then
				puts(nx,sx)
				ix=machine_func(20,fn)
			end if
			sx=""
		elsif fc!='\r' then
			sx&=fc
		end if
	end while
	close(nx)
	close(fn)
	system("del "&tsv_file,2)
	system("rename xxx.tsv "&tsv_file,2)
	fn=open(tsv_file,TSV_FILES[fx][TSV_MODE])
	if fn>0 then
		TSV_FILES[fx][TSV_NR]=fn
	else
		fx=tsv_close(tsv_file)
	end if
	-- fx=machine_func(19,{fn,0})
	return ix
end function

--***************************************************************************
-- These routines report on records
--***************************************************************************

global function tsv_rec_filter(sequence tsv_file, object filter_id)
-- your filter_id is a rountine_id reference to a function
-- that checks a sequence of words and returns 0 or +1
	sequence sx, ax
	integer ix
	-- you can not read from a file that is not open
	fx=tsv_file_check(tsv_file)
	if fx<1 then
		return TSV_ERROR_OPEN
	end if
	-- it is not wise to read a record if it is not in read mode
	if equal(TSV_FILES[fx][TSV_MODE],READ)=0
	and equal(TSV_FILES[fx][TSV_MODE],UPDATE)=0
	then
		return TSV_ERROR_MODE
	end if
	-- let's start reading
	ax={}
	sx={""}
	ix='f'	-- not needed line
	fn=TSV_FILES[fx][TSV_NR]
	while 1 do
		fc=getc(fn)
		if fc='\t' then
			sx=append(sx,"")
		elsif fc='\n' then
			if ix='y' then
				if sequence(filter_id) 
				or call_func(filter_id,{sx})=1 then
					ax=append(ax,sx)
				end if
			else
				ix='y'
			end if
			sx={""}
		elsif fc=-1 then
			exit
		elsif fc!='\r' then
			sx[$]&=fc
		end if
	end while
	-- after the last line there is no \n
	if ix='y' then
		if sequence(filter_id)
		or call_func(filter_id,{sx})=1 then
			ax=append(ax,sx)
		end if
	end if
	fx=machine_func(19,{fn,0})
	return ax
end function

global function tsv_rec_all(sequence tsv_file)
-- think twice!
-- to load all records might require a lot of memory
	return tsv_rec_filter(tsv_file,"all")
end function

global function tsv_rec_get(sequence tsv_file, sequence tsv_key)
	sequence sx
	integer ix
	-- you can not read from a file that is not open
	fx=tsv_file_check(tsv_file)
	if fx<1 then
		return TSV_ERROR_OPEN
	end if
	-- it is not wise to read a record if it is not in read mode
	if equal(TSV_FILES[fx][TSV_MODE],READ)=0
	and equal(TSV_FILES[fx][TSV_MODE],UPDATE)=0
	then
		return TSV_ERROR_MODE
	end if
	-- let's start reading
	sx={""}
	ix='n'	-- not the needed line
	fn=TSV_FILES[fx][TSV_NR]
	while 1 do
		fc=getc(fn)
		if fc='\t' then
			if equal(sx[KEY],tsv_key) then
				sx=append(sx,"")
			else
				ix='n'
			end if
		elsif fc='\n' then
			if equal(sx[KEY],tsv_key) then
				exit
			else
				sx={""}
				ix='y'
			end if
		elsif fc=-1 then
			exit
		elsif ix='y' then
			if fc!='\r' then
				sx[$]&=fc
			end if
		end if
	end while
	fx=machine_func(19,{fn,0})
	if equal(sx[KEY],tsv_key) then
		return sx
	else
		return TSV_ERROR_FIND
	end if
end function

global function tsv_rec_key(sequence tsv_file, sequence tsv_key)
	sequence sx
	integer ix
	-- you can not read from a file that is not open
	fx=tsv_file_check(tsv_file)
	if fx<1 then
		return TSV_ERROR_OPEN
	end if
	-- it is not wise to read a record if it is not in read mode
	if equal(TSV_FILES[fx][TSV_MODE],READ)=0
	and equal(TSV_FILES[fx][TSV_MODE],UPDATE)=0
	then
		return TSV_ERROR_MODE
	end if
	-- let's start reading
	sx=""
	ix='f'	-- not needed line
	fn=TSV_FILES[fx][TSV_NR]
	while 1 do
		fc=getc(fn)
		if fc=-1 then
			exit
		elsif fc='\n' then
			sx=""
			ix='y'
		elsif ix='y' then
			if fc='\t' then
				ix='n'
				if equal(sx,tsv_key) then
					fx=machine_func(19,{fn,0})
					return TSV_OK
				end if
			elsif fc!='\r' then
				sx&=fc
			end if
		end if
	end while
	fx=machine_func(19,{fn,0})
	return TSV_ERROR_FIND
end function

global function tsv_rec_nr(sequence tsv_file, integer tsv_line)
	sequence sx
	integer ix,cx
	-- you can not read from a file that is not open
	fx=tsv_file_check(tsv_file)
	if fx<1 then
		return TSV_ERROR_OPEN
	end if
	-- it is not wise to read a record if it is not in read mode
	if equal(TSV_FILES[fx][TSV_MODE],READ)=0
	and equal(TSV_FILES[fx][TSV_MODE],UPDATE)=0
	then
		return TSV_ERROR_MODE
	end if
	-- let's start reading
	sx={""}
	ix=0	-- not needed line (the first line)
	fn=TSV_FILES[fx][TSV_NR]
	while 1 do
		fc=getc(fn)
		if fc=-1 then
			exit
		elsif fc='\n' then
			-- sx={""}
			ix+=1
		elsif ix=tsv_line then
			if fc='\t' then
				sx=append(sx,"")
			elsif fc!='\r' then
				sx[$]&=fc
			end if
		elsif ix>tsv_line then
			exit
		end if
	end while
	fx=machine_func(19,{fn,0})
	if equal(sx,{""}) then
	       	return TSV_ERROR_FIND
	else
		return sx
	end if
end function
--***************************************************************************
